with Universal_Types;
use Universal_Types;
with Universal_Constants;
with Sqrt;

package body Orbit is

    subtype Acceleration is Numeric;

    procedure Compute_Acceleration 
		 (C : Coordinates_2D.Object; Ax, Ay : out Acceleration) is

	R : Numeric;
	R2 : Numeric;

	X : Coordinates_2D.Position := Coordinates_2D.X_Position (C);
	Y : Coordinates_2D.Position := Coordinates_2D.Y_Position (C);

	Temporary_1 : Numeric;
	Temporary_2 : Numeric;
	Ax_T, Ay_T : Numeric;
	Sf_1 : constant Numeric := 30_000;
	Sf_2 : constant Numeric := 100;
	Sf_3 : constant Numeric := Sf_1 / (Sf_2 * 100);

	--
	-- Do not tinker with this code in an attempt to "optimize" it
	-- Operations are done in very special order to prevent overflow
	-- and to maintain the maximum number of significant digits
	--
    begin

	X := X / Universal_Constants.Scale_Factor;
	Y := Y / Universal_Constants.Scale_Factor;
	R2 := (Y * Y) + (X * X);
	R := Sqrt (R2);
	--
	Temporary_1 := Sf_1 / R;
	Temporary_2 := R2 / Sf_2;
	--
	Ax_T := Temporary_1 * X;
	Ax_T := Ax_T / Temporary_2;
	Ax_T := Ax_T / Sf_3;
	Ax := Ax_T * Universal_Constants.Gravitational_Constant;
	--
	Ay_T := Temporary_1 * Y;
	Ay_T := Ay_T / Temporary_2;
	Ay_T := Ay_T / Sf_3;
	Ay := Ay_T * Universal_Constants.Gravitational_Constant;
	-- Accelerations are 100x bigger than normal, on purpose
	-- This matches with scale_factor in state_variable
	--
    end Compute_Acceleration;

    function Make (Coordinate : Coordinates_2D.Object; 
		   Velocity : Velocity_2D.Object; 
		   Initialize_Orbit : Boolean := False) return Object is
	Two : constant Numeric := 2; -- user for half step calulation below
	X_Offset : Velocity_2D.Offset;
	Y_Offset : Velocity_2D.Offset;
	Delta_Time : Velocity_2D.Time;
	X_Acceleration : Numeric;
	Y_Acceleration : Numeric;
    begin

	-- do Feynman half step ion accelleration to improve accuracy
	if Initialize_Orbit then
	    X_Offset := Velocity_2D.X_Offset (Velocity);
	    Y_Offset := Velocity_2D.Y_Offset (Velocity);
	    Delta_Time := Velocity_2D.Delta_Time (Velocity);
	    Compute_Acceleration (Coordinate, X_Acceleration, Y_Acceleration);
	    X_Offset := 
	       ((X_Acceleration * Delta_Time) + (X_Offset * Two)) / Two;
	    Y_Offset := 
	       ((Y_Acceleration * Delta_Time) + (Y_Offset * Two)) / Two;
	    return (C => Coordinate, 
		    V => Velocity_2D.Make (X_Offset, Y_Offset, Delta_Time));
	else
	    return (C => Coordinate, V => Velocity);
	end if;
    end Make;

    function Coordinate (O : Object) return Coordinates_2D.Object is
    begin
	return O.C;
    end Coordinate;

    function Velocity (O : Object) return Velocity_2D.Object is
    begin
	return O.V;
    end Velocity;

    procedure Euler_Update (O : in out Object) is

	Current_Coordinates : Coordinates_2D.Object := Coordinate (O);
	X : Coordinates_2D.Position := 
	   Coordinates_2D.X_Position (Current_Coordinates);
	Y : Coordinates_2D.Position := 
	   Coordinates_2D.Y_Position (Current_Coordinates);

	Current_Velocity : Velocity_2D.Object := Velocity (O);
	X_Offset : Velocity_2D.Offset := 
	   Velocity_2D.X_Offset (Current_Velocity);
	Y_Offset : Velocity_2D.Offset := 
	   Velocity_2D.Y_Offset (Current_Velocity);
	Delta_Time : Velocity_2D.Offset := 
	   Velocity_2D.Delta_Time (Current_Velocity);

	X_Acceleration : Numeric;
	Y_Acceleration : Numeric;
    begin

	-- a simple Euler integration to update the state

	X := X + X_Offset * Delta_Time;
	Y := Y + Y_Offset * Delta_Time;
	Current_Coordinates := Coordinates_2D.Make (X, Y);
	-- compute new coordiantes

	Compute_Acceleration (Current_Coordinates, 
			      X_Acceleration, Y_Acceleration);
	X_Offset := X_Offset + X_Acceleration * Delta_Time;
	Y_Offset := Y_Offset + Y_Acceleration * Delta_Time;
	Current_Velocity := Velocity_2D.Make (X_Offset, Y_Offset, Delta_Time);
	-- compute updates obital velocity

	O := Make (Current_Coordinates, Current_Velocity);
	-- update orbit   

    end Euler_Update;

end Orbit;

